/*
	Magic number handling.

	This class loads, stores and then searches the magic numbers.
	
	Author: Matthew Allen <fret@memecode.com>
	Date: 5/12/2003
*/


#include "mime-types-priv.h"
#include "magic.h"
#define min(a, b) ((a)<(b)?(a):(b))

class Number
{
public:
	int Indent;
	int Offset;
	int Length;
	uchar *Value;
	uchar *Mask;
	int WordSize;
	int Range;
	
	Number()
	{
		Indent = 0;
		WordSize = 1;
		Offset = 0;
		Length = 0;
		Value = 0;
		Mask = 0;
		Range = 1;
	}
	
	~Number()
	{
		DeleteArray(Value);
		DeleteArray(Mask);
	}
	
	int ReadNumber(char *&s)
	{
		int i = atoi(s);
		while (isdigit(*s)) s++;
		return i;
	}
	
	uchar *ReadBytes(char *&s)
	{
		uchar *Buf = new uchar[Length];
		if (Buf)
		{
			memcpy(Buf, s, Length);
			s += Length;
		}
		return Buf;
	}
	
	uint WordAt(uchar *Data, int Index)
	{
		switch (WordSize)
		{
			case 1:
			{
				return (uchar)Data[Index];
			}
			case 2:
			{
				uint16 *p = (uint16*)Data;
				return	((p[Index] & 0xff00) >> 8) |
						((p[Index] & 0x00ff) << 8);
			}
			case 4:
			{
				uint32 *p = (uint32*)Data;
				return	((p[Index] & 0xff000000) >> 24) |
						((p[Index] & 0x00ff0000) >> 8) |
						((p[Index] & 0x0000ff00) << 8) |
						((p[Index] & 0x000000ff) << 24);
			}
			default:
			{
				printf("%s:%i - Unknown word size %i.\n", __FILE__, __LINE__, WordSize);
				assert(0);
			}
		}
		
		return 0;
	}

	#define IsPrintable(c)		((c)>=' ' && (c)<0x7f)

	bool Match(char *Data, int Len, bool Debug)
	{
		bool Status = false;
		
		if (Value)
		{
			if (Debug)
			{
				char In[32];
				memset(In, 0, sizeof(In));
				memset(In, ' ', (Indent+1)<<2);

				printf("%sNumber, indent=%i, data is %i bytes, from %i to %i, len=%i, wordsize=%i, value=",
					In, Indent, Len, Offset, Offset+Range, Length, WordSize);
				for (int i=0; i<Length; i++)
				{
					if (IsPrintable(Value[i]))
					{
						int n = 0;
						while (IsPrintable(Value[i+n]) && (i+n)<Length)
						{
							n++;
						}
						printf("'%.*s', ", n, Value+i);
						i += n - 1;
					}
					else
					{
						printf("0x%02.2X, ", (uchar)Value[i]);
					}
				}
			}
		
			for (int i = Offset; i < (Len - Length) && i < Offset + Range; i++)
			{
				int Words = Length / WordSize;
				if (WordSize > 1) assert(WordSize % Length == 0);
				int DefMask = 0;
				switch (WordSize)
				{
					default:
					case 1: DefMask = 0xff; break;
					case 2: DefMask = 0xffff; break;
					case 3: DefMask = 0xffffff; break;
					case 4: DefMask = 0xffffffff; break;
				}
				
				int n;
				for (n=0; n<Words; n++)
				{
					uint Msk = Mask ? WordAt(Mask, n) : DefMask;
					uint Dat = WordAt((uchar*)Data+i, n);
					uint Val = WordAt(Value, n);

					if (Debug) printf("\t\t[%i of %i] Dat=%x Msk=%x Val=%x\n", n, Words, Dat, Msk, Val);
					
					if ((Dat & Msk) != (Val & Msk))
					{
						break;
					}
				}
				if (n == Words)
				{
					if (Debug)
					{
						printf(" = MATCH!\n");
					}
					return true;
				}
			}
		}
		
		if (Debug) printf(" = miss\n");
		return Status;
	}
};

class Magic
{
public:
	int Priority;
	char *Type;
	Array<Number*> Numbers;
	
	Magic()
	{
		Priority = 0;
		Type = 0;
	}
	
	~Magic()
	{
		DeleteArray(Type);
		for (int i=0; i<Numbers.Length(); i++)
		{
			DeleteObj(Numbers[i]);
		}
	}

	bool _Match(bool ParentStatus, char *Data, int Len, bool Debug, int Level, int &Index)
	{
		bool Status = false;
		char In[32];
		memset(In, 0, sizeof(In));
		memset(In, ' ', (Level+1)<<2);
		
		if (Debug)
		{
			printf("%s\n", Type);
		}

		while (Index < Numbers.Length())
		{
			Number *n = Numbers[Index];
			
			if (n)
			{
				bool s = Status;
				if (n->Indent < Level)
				{
					break;
				}
				else
				{
					if (n->Indent != Level)
					{
						printf("Level mismatch: Indent=%i Level=%i\n", n->Indent, Level);
						assert(0);
					}
					
					// Check current
					bool r = n->Match(Data, Len, Debug);
					Index++;

					// Has children?
					Number *Next = Numbers[Index];
					if (Next && Next->Indent > Level)
					{
						// Resolve child numbers
						bool c = _Match(Status, Data, Len, Debug, Next->Indent, Index);
						if (Debug) printf("%sAnd (child): %i && %i = %i\n", In, c, r, c && r);
						r = c && r;
					}

					if (Debug) printf("%sOr: %i || %i = %i\n", In, Status, r, Status || r);
					Status = Status || r;

				}
			}
			else break;
		}
	
		return Status;
	}

	bool Match(char *Data, int Len, bool Debug = false)
	{
		/* Data[1] == 'E' &&
					(strstr(Type, "x-exe") != 0 || strstr(Type, "x-core") != 0);
		if (Debug)
		{
			printf("Checking for '%s', with %i numbers.\n", Type, Numbers.Length());
		}
		*/
		
		int Index = 0;
		return _Match(true, Data, Len, Debug, 0, Index);
	}
};

class MagicLutPrivate
{
public:
	Array<Magic*> Magics;
};

MagicLut::MagicLut()
{
	d = new MagicLutPrivate;
}

MagicLut::~MagicLut()
{
	DeleteObj(d);
}

bool MagicLut::Load(char *File)
{
	bool Status = false;

	char *Txt = ReadTextFile(File);
	if (Txt)
	{
		// Check for file header...
		if (strncmp(Txt, "MIME-Magic", 10) == 0 &&
			Txt[10] == 0 &&
			Txt[11] == '\n')
		{
			// Read in magic number defs...
			for (char *s = Txt + 12; *s == '['; )
			{
				// Read magic type header
				char *Num = ++s;
				s = strchr(s, ':');
				if (!s)
				{
					printf("%s:%i - Missing ':' in type def.\n", __FILE__, __LINE__);
					break;
				}
				
				Magic *m = new Magic;
				if (!m)
				{
					break;
				}
				
				// Priority
				m->Priority = atoi(Num);
				char *Name = ++s;
				s = strchr(s, ']');
				if (!s)
				{
					printf("%s:%i - Missing ':' in type def.\n", __FILE__, __LINE__);
					break;
				}
				
				// Mime-type
				m->Type = NewStr(Name, (int)s-(int)Name);
				s = strchr(s, '\n');
				if (!s)
				{
					printf("%s:%i - Missing '\n' after type def.\n", __FILE__, __LINE__);
					break;
				}
				s++;
				
				// Read numbers
				while (isdigit(*s) || *s == '>')
				{
					Number *n = new Number;
					if (!n)
					{
						break;
					}

					if (isdigit(*s))
					{
						// Has indent
						n->Indent = atoi(s);
						while (isdigit(*s)) s++;
					}
					
					if (*s != '>')
					{
						printf("%s:%i - Missing '>' in number def.\n", __FILE__, __LINE__);
						break;
					}
					
					s++;
					n->Offset = n->ReadNumber(s);
					
					if (*s != '=')
					{
						printf("%s:%i - Missing '=' in number def.\n", __FILE__, __LINE__);
						break;
					}
					s++;

					// read the binary length
					n->Length = (uchar)s[0] << 8 | (uchar)s[1];
					s += 2;
					
					// read the value
					n->Value = n->ReadBytes(s);
					
					// check for mask
					if (*s == '&')
					{
						s++;
						n->Mask = n->ReadBytes(s);
					}
					
					// check for word-size
					if (*s == '~')
					{
						s++;
						n->WordSize = n->ReadNumber(s);
					}
					
					// check for range-length
					if (*s == '+')
					{
						s++;
						n->Range = n->ReadNumber(s);
					}
					
					// Insert the number into the list
					m->Numbers[m->Numbers.Length()] = n;
					
					if (*s != '\n')
					{
						printf("%s:%i - Missing '\n' at end of number def.\n", __FILE__, __LINE__);
						break;
					}
					s++;
				}
				
				// Store the magic
				d->Magics[d->Magics.Length()] = m;
			}
		}
		
		DeleteArray(Txt);
	}
	
	return Status;
}

char *MagicLut::GetType(char *Data, int Len, char *Hint, bool Debug)
{
	if (Data && Len > 0)
	{
		Magic *Best = 0;
		int i;

		if (Debug) printf("Hint=%s\n", Hint);
		
		if (Hint)
		{
			// Check the hint for accuracy.
			for (i=0; i<d->Magics.Length(); i++)
			{
				Magic *m = d->Magics[i];
				if (m && m->Type)
				{
					if (stricmp(m->Type, Hint) == 0)
					{
						// Test the hinted type
						printf("Matching hint '%s'... ", m->Type);
						if (m->Match(Data, Len))
						{
							printf("matched.\n");
							return NewStr(m->Type);
						}

						printf("failed.\n");
						break;
					}
				}
			}
			
			if (i >= d->Magics.Length())
			{
				// Lets assume that the hinted type is correct if we can't
				// test it for sure.
				return NewStr(Hint);
			}
		}
		
		if (Debug) printf("Magics=%i\n", d->Magics.Length());

		// Do a full check of all the magics
		for (i=0; i<d->Magics.Length(); i++)
		{
			Magic *m = d->Magics[i];
			if (m)
			{
				if (m->Match(Data, Len, Debug))
				{
					if (!Best || Best->Priority < m->Priority)
					{
						Best = m;
					}
				}
			}
		}
		
		if (Best)
		{
			return NewStr(Best->Type);
		}
	}
	
	return 0;
}

